package device

import (
	"crypto/rand"
	"encoding/hex"
	"io"
	"log"
	"net"
	"net/http"
	"strings"

	"git.autistici.org/id/auth"
	"github.com/gorilla/securecookie"
	"github.com/mssola/user_agent"
)

const (
	deviceIDCookieName   = "_dev"
	deviceIDCookieMaxAge = 10 * 365 * 86400
)

type DeviceID []byte

func (d DeviceID) String() string {
	return hex.EncodeToString(d)
}

func randomDeviceID() DeviceID {
	b := make([]byte, 8)
	if _, err := io.ReadFull(rand.Reader, b[:]); err != nil {
		panic(err)
	}
	return b
}

// Manager can provide DeviceInfo entries for incoming HTTP requests.
type Manager struct {
	sc        *securecookie.SecureCookie
	urlPrefix string
	geodb     *geoIPDb
}

// Config stores options for the device info manager.
type Config struct {
	AuthKey        string   `yaml:"auth_key"`
	GeoIPDataFiles []string `yaml:"geo_ip_data_files"`
}

// New returns a new Manager with the given configuration.
func New(config *Config, urlPrefix string) (*Manager, error) {
	if config == nil {
		config = &Config{}
	}

	geodb, err := newGeoIP(config.GeoIPDataFiles)
	if err != nil {
		log.Printf("Warning: GeoIP disabled: %v", err)
	}

	// This should only happen in tests.
	if config.AuthKey == "" {
		log.Printf("Warning: device_manager.auth_key unset, generating temporary random secrets")
		config.AuthKey = string(securecookie.GenerateRandomKey(64))
	}

	sc := securecookie.New([]byte(config.AuthKey), nil)
	sc.MaxAge(deviceIDCookieMaxAge)
	sc.SetSerializer(securecookie.NopEncoder{})

	return &Manager{
		sc:        sc,
		geodb:     geodb,
		urlPrefix: urlPrefix,
	}, nil
}

// GetDeviceInfoFromRequest will retrieve or create a DeviceInfo
// object for the given request. It will always return a valid object.
// The ResponseWriter is needed to store the unique ID on the client
// when a new device info object is created.
func (m *Manager) GetDeviceInfoFromRequest(w http.ResponseWriter, req *http.Request) *auth.DeviceInfo {
	devID, ok := m.getDeviceCookie(req)
	if !ok || len(devID) == 0 {
		// Generate a new Device ID and save it on the client.
		devID = randomDeviceID()
		if err := m.setDeviceCookie(w, devID); err != nil {
			// This is likely a misconfiguration issue, so
			// we want to know about it.
			log.Printf("error saving device manager session: %v", err)
		}
	}

	uaStr := req.UserAgent()
	ua := user_agent.New(uaStr)
	browser, _ := ua.Browser()
	d := auth.DeviceInfo{
		ID:        devID.String(),
		UserAgent: uaStr,
		Mobile:    ua.Mobile(),
		OS:        ua.OS(),
		Browser:   browser,
	}

	// Special check for .onion HTTP hosts - sets the zone to
	// "onion" and skips the IP-based checks.
	if req.Host != "" && strings.HasSuffix(req.Host, ".onion") {
		d.RemoteZone = "onion"
	} else {
		// If we have an IP address, we can add more information.
		if ip := getIPFromRequest(req); ip != nil {
			d.RemoteAddr = ip.String()

			// Country lookup.
			if m.geodb != nil {
				if zone, err := m.geodb.getZoneForIP(ip); err == nil {
					d.RemoteZone = zone
				}
			}
		}
	}

	return &d
}

func (m *Manager) getDeviceCookie(r *http.Request) (DeviceID, bool) {
	if cookie, err := r.Cookie(deviceIDCookieName); err == nil {
		var value []byte
		if err = m.sc.Decode(deviceIDCookieName, cookie.Value, &value); err == nil {
			return DeviceID(value), true
		}
	}
	return nil, false
}

func (m *Manager) setDeviceCookie(w http.ResponseWriter, value DeviceID) error {
	encoded, err := m.sc.Encode(deviceIDCookieName, []byte(value))
	if err != nil {
		return err
	}
	cookie := &http.Cookie{
		Name:     deviceIDCookieName,
		Value:    encoded,
		Path:     m.urlPrefix + "/",
		Secure:   true,
		HttpOnly: true,
		MaxAge:   deviceIDCookieMaxAge,
	}
	http.SetCookie(w, cookie)
	return nil
}

func getIPFromRequest(req *http.Request) net.IP {
	// Parse the RemoteAddr Request field, for starters.
	host, _, err := net.SplitHostPort(req.RemoteAddr)
	if err != nil {
		host = req.RemoteAddr
	}
	return net.ParseIP(host)
}
