package server

import (
	"io/ioutil"
	"net/http"
	"net/http/httptest"
	"net/url"
	"os"
	"path/filepath"
	"strings"
	"testing"

	"git.autistici.org/id/go-sso/httpsso"
	"github.com/gorilla/securecookie"
)

// Create a SSO-wrapped service.
func createTestProtectedService(t testing.TB, serverURL, tmpdir string) *httptest.Server {
	ssoPubKey, err := ioutil.ReadFile(filepath.Join(tmpdir, "public"))
	if err != nil {
		t.Fatalf("oops, can't read sso public key: %v", err)
	}

	w, err := httpsso.NewSSOWrapper(
		serverURL,
		ssoPubKey,
		"example.com",
		securecookie.GenerateRandomKey(64),
		securecookie.GenerateRandomKey(32),
		0,
	)
	if err != nil {
		t.Fatalf("NewSSOWrapper(): %v", err)
	}

	h := w.Wrap(http.HandlerFunc(func(w http.ResponseWriter, _ *http.Request) {
		w.Write([]byte("OK")) // nolint
	}), "service.example.com/", nil)

	return httptest.NewTLSServer(h)
}

func startTestHTTPServerAndApp(t testing.TB) (string, *httptest.Server, *httptest.Server) {
	tmpdir, _ := ioutil.TempDir("", "")
	config := testConfig(t, tmpdir, "")
	srv := createTestHTTPServer(t, config)
	app := createTestProtectedService(t, "https://login.example.com/", tmpdir)
	return tmpdir, srv, app
}

func checkIsProtectedService(t testing.TB, resp *http.Response) {
	data, err := ioutil.ReadAll(resp.Body)
	if err != nil {
		t.Fatalf("reading body: %v", err)
	}
	if s := string(data); s != "OK" {
		t.Fatalf("not the target application, response body='%s'", s)
	}
}

func checkLogoutPageHasLinks(t testing.TB, resp *http.Response) {
	data, err := ioutil.ReadAll(resp.Body)
	if err != nil {
		t.Fatalf("reading body: %v", err)
	}
	logoutURL := "https://service.example.com/sso_logout"
	if sdata := string(data); !strings.Contains(sdata, logoutURL) {
		t.Fatalf("service logout URL not found in logout page:\n%s", sdata)
	}
}

func addrFromURL(s string) string {
	u, _ := url.Parse(s)
	if !strings.Contains(u.Host, ":") {
		return u.Host + ":443"
	}
	return u.Host
}

// The integration test spins up an actual service and verifies the
// interaction between it and the login application, using DNS-level
// overrides to ensure proper name validation.
func TestIntegration(t *testing.T) {
	tmpdir, srv, app := startTestHTTPServerAndApp(t)
	defer os.RemoveAll(tmpdir)
	defer srv.Close()
	defer app.Close()

	c := makeHTTPClient(map[string]string{
		"login.example.com:443":   addrFromURL(srv.URL),
		"service.example.com:443": addrFromURL(app.URL),
	}, true)

	doGet(t, c, "https://service.example.com/", checkStatusOk, checkLoginPasswordPage)

	v := make(url.Values)
	v.Set("username", "testuser")
	v.Set("password", "password")
	doPostForm(t, c, "https://login.example.com/login", v, checkStatusOk, checkIsProtectedService)

	// Now attempt to logout, and verify that we can't access the service anymore.
	doGet(t, c, "https://login.example.com/logout", checkStatusOk, checkLogoutPageHasLinks)
	doGet(t, c, "https://service.example.com/sso_logout", checkStatusOk)
	doGet(t, c, "https://service.example.com/", checkStatusOk, checkLoginPasswordPage)
}
